var parser = require("./parser");

function combine(template, context) {
  if(!(this instanceof combine)) {
    return new combine(template, context);
  }

  this.template = template;
  this.context = context || {};

  // TODO: Consolidate filters and partials

  // Filters are specific to each template, decided to isolate from prototype
  // until I can figure out a more elegant solution.
  var _filters = {};
  this.filters = {

    // Create/Update
    add: function(name, cb) {
      _filters[name] = _wrapFilter(cb);
    },

    // Read
    get: function(name) {
      return _filters[name];
    },

    // Delete
    remove: function(name) {
      delete _filters[name];
    }

  };

  // Partials are specific to each template, decided to isolate from prototype
  // until I can figure out a more elegant solution.
  var _partials = {};
  this.partials = {

    // Create/Update
    add: function(name, template, context) {
      _partials[name] = { template: template, context: context };
    },

    // Read
    get: function(name) {
      return _partials[name];
    },

    // Delete
    remove: function(name) {
      delete _partials[name];
    }

  };
}

combine.version = "0.0.1";
combine.prototype = {

  render: function() {
    var self = this;

    // Missing template or context... can't exactly do anything here...
    if (!self.template || !self.context) {
      throw new Error("Missing template or context");
    } 

    // TODO: Run pre-render hooks

    // Override delimiters
    self.delimiters = self.delimiters || {};
    self.delimiters.__proto__ = {
      START_PROP: "{{",
      END_PROP: "}}",
      START_EXPR: "{%",
      END_EXPR: "%}",
      COMMENT: "--",
      FILTER: "|"
    };

    // Actually render
    var output = parser(self, self.template, self.context, self.delimiters);

    // TODO: Run post render hooks

    return output;
  },

  delimiters: {}
};

function _wrapFilter(func) {
  var self = this;
  var context;

  return function() {
    context = context ? context : { original: self.template };

    return func.apply(context, arguments);
  };
}

module.exports = combine;
