var tokenizer = function(_tokens) {
  return function(template, stack, tokens) {
    tokens = tokens || _tokens;
    _tokens = tokens;

    var captures, token;
    var names = Object.keys(tokens);
    var len = names.length;
    var name = '';
    var i = 0;

    while(template.length) {
      for (i = 0; i<len; i++) {
        name = names[i];

        if (captures = tokens[name].exec(template)) {
          token = { name: name, captures: captures };
          template = template.replace(tokens[name], '');

          if (!token) {
            throw new Error("Unable to parse");
          }

          if (captures[0]) {
            stack.push(token);
          }
        }
      }
    }
  };
}();

var parse = function() {
  var obj, innerText, names;
  var output = '';
  var retVal = '';
  var _delimiters;

  function runFilter(self, method, args, obj) {
    var filter = self.filters.get(method);   

    if (filter == null) {
      throw new Error("Filter " + method + " not found");
    }

    args.unshift(obj);

    return filter.apply(obj, args);
  }

  function runConditional(self, args, context, mode) {
    var i, arg, left, right, comparator;
    var len = args.length;
    var truthy = true;
    var not;
    var stringExp = /\'|\"/gi;
    var compare = {
      "==": function(left, right) {
        if (left == right) {
          return true;
        }

        return false;
      },
      "!=": function(left, right) {
        if (left != right) {
          return true;
        }

        return false;
      },
      ">": function(left, right) {
        if (left > right) {
          return true;
        }

        return false;
      },
      ">=": function(left, right) {
        if (this["=="](left, right) || this[">"](left, right)) {
          return true;
        }

        return false;
      },
      "<": function(left, right) {
        if (left < right) {
          return true;
        }

        return false;
      },
      "<=": function(left, right) {
        if (this["=="](left, right) || this["<"](left, right)) {
          return true;
        }

        return false;
      },
      "not": function(val) {
        return !val;
      }
    };

    for (i=0; i<len; i++) {
      // Normalize numbers and everything else to strings
      arg = !global.isNaN(args[i]) ? +args[i] : args[i]+"";

      // Set not
      if (arg === "not") {
        not = true;
        continue;
      }

      // Set conditional
      if (compare[arg]) {
        comparator = arg;
        continue;
      }

      // Check for string
      if (stringExp.exec(arg)) {
        arg = arg.replace(stringExp, '');
      }
      else if (typeof arg === "string") {
        arg = context[arg] || undefined;
      }

      // Negative this value
      if (not) {
        not = undefined;

        if (left === undefined) {
          left = compare.not(arg);
          continue;
        }
        right = arg;
        continue;
      }

      if (left === undefined) {
        left = arg;
        continue;
      }
      right = arg;
      break;
    }

    if (comparator) {
      if (compare[comparator]) {
        truthy = compare[comparator](left, right);
      }
      else {
        truthy = false;
      }
    }
    else {
      truthy = left;
    }

    return truthy;
  }

  function runPartial(self, method, args, obj) {
    var _output, innerText;
    var stack = [];
    var name = args[0];
    var partial = self.partials.get(name);

    if(partial == null) {
      throw new Error("Partial " + name + " not found");
    }

    if (error = tokenizer(partial.template, stack)) {
      throw new Error(error);
    }

    stack.reverse();

    // Save original output state
    _output = output;

    // Clear it for the inner text
    output = "";

    innerText = parse(self, partial.context, stack, _delimiters);

    output = _output;

    return innerText;
  }

  var mode = function() {
    var modes = [];

    return {
      set: function(val) {
        modes = modes.concat(val);
      },
      get: function() {
        return modes;
      },
      unset: function(val) {
        modes.splice(modes.indexOf(val), 1);
      },
      exists: function(val) {
        return ~modes.indexOf(val);
      },
      clear: function() {
        modes = [];
      },
      count: function(val) {
        // Potential sexier way
        // for (var count = 0, lastIndex = 0; ~( lastIndex = modes.indexOf(val, lastIndex) ); count++);
        var count = 0;
        for (var i=0; i<modes.length; i++) {
          if (modes[i] === val) {
            count = count + 1;
          }
        }

        return count;
      }
    };
  }();

  var loop = function() {
    var _mode, key, keys, stack, error, context;
    var loops = [];

    return {
      execArray: function(self, array, oldContext) {
        var i, iLen;
        _mode = mode.get();
        mode.unset('loop');

        if (array && array.length) {
          for (i=0, iLen=array.length; i<iLen; i++) {
            stack = [];

            if (error = tokenizer(innerText, stack)) {
              throw new Error(error);
            }

            stack.reverse();

            // Ensure these required properties cannot be overwritten
            context = { '.': array[i], i: i, length: iLen, original: array };
            if (typeof array[i] === 'object') {
              context.__proto__ = array[i];
              array[i].__proto__ = oldContext;
            }
            else {
              context.__proto__ = oldContext;
            }

            output = parse(self, context, stack, _delimiters);
          }
        }
        mode.set(_mode);

        innerText = output;
      },

      execObj: function(self, obj, name, oldContext) {
        _mode = mode.get();
        mode.unset('loop');

        keys = Object.keys(obj);
        if (keys.length) {
          for (i=0, iLen=keys.length; i<iLen; i++) {
            key = keys[i];
            stack = [];

            if (error = tokenizer(innerText, stack)) {
              throw new Error(error);
            }

            stack.reverse();

            // Ensure these required properties cannot be overwritten
            // FIXME: Saner defaults for object keys
            context = { original: obj };
            context[name.key || '.'] = key;
            context[name.obj || 'original'] = obj;
            context[name.val || '$'] = obj[key];
            if (typeof obj[key] === 'object') {
              context.__proto__ = obj[key];
              obj[key].__proto__ = oldContext;
            }
            else {
              context.__proto__ = oldContext;
            }
          
            output = parse(self, context, stack, _delimiters);
          }
        }
        mode.set(_mode);
      },

      clear: function() {
        loops = [];
      }
    };
  }();

  return function(self, context, stack, delimiters) {
    var tmpVar, method, args;
    var token = stack.pop();
    var capture = token.captures[0];

    // Cache for later use
    _delimiters = delimiters || _delimiters;

    switch(token.name) {
      case "START_PROP":
        if (mode.exists("skip")) {
          break;
        }

        if (mode.exists("loop")) {
          innerText += capture;
          break;
        }

        mode.set("prop");

        break;

      case "END_PROP":
        if (mode.exists("skip")) {
          break;
        }

        if (mode.exists("loop")) {
          innerText += capture;
          break;
        }

        // Object is in context
        if (typeof obj !== 'object') {
          output += obj;
        }
        // Keep brackets
        else {
          output += delimiters.START_PROP + obj.prop + delimiters.END_PROP;
        }

        mode.unset("prop");

        break;

      case "START_EXPR":
        mode.set("expr");

        if (mode.exists("skip")) {
          break;
        }

        if (mode.count("loop") > 1) {
          innerText += capture;
        }

        break;

      case "END_EXPR":
        mode.unset("expr");

        if (mode.exists("skip")) {
          if (obj === "--") {
            mode.unset("skip");
          }

          break;
        }

        if (mode.exists("partial")) {
          output += obj;
          mode.unset("partial");
        }

        if (obj === "endeach") {
          mode.unset("loop");
        }

        if (mode.count("loop") > 1) {
          innerText += capture;

        }

        break;

      case "STRING":
        if (mode.exists("prop")) {
          if (mode.exists("skip")) {
            break;
          }

          if (mode.exists("loop")) {
            innerText += capture;
            break;
          }

          args = capture.split('.');
          if (args.length > 1 && context[args[0]]) { 
            tmpVar = context[args[0]][args[1]];
            // TODO: Need recursion here
            if (args[1] in context[args[0]]) {
              // Allow for function or literal values
              obj = typeof tmpVar == "function" ? tmpVar() : tmpVar;
              // Force toString()
              obj = obj.toString();
            }
            else {
              obj = { prop: capture };
            }
          }
          else {
            if (capture in context) {
              // Allow for function or literal values
              obj = typeof context[capture] == "function" ? context[capture]() : context[capture];
              // Force toString()
              obj = obj.toString();
            }
            else {
              obj = { prop: capture };
            }
          }
        }
        else if (mode.exists("filter")) {
          if (mode.exists("skip")) {
            break;
          }

          if (mode.exists("loop")) {
            innerText += capture;
            break;
          }

          tmpVar = capture.split(' ');
          method = tmpVar.shift();
          args = tmpVar;

          obj = runFilter(self, method, args, obj);
          mode.unset("filter");
          mode.set("string");
        }
        else if (mode.exists("expr")) {
          tmpVar = capture.split(' ');
          method = tmpVar.shift();
          args = tmpVar;

          if (method === "if") {
            if (mode.exists("else")) {
              mode.unset("else");
            }

            if (mode.exists("skip")) {
              break;
            }

            if (mode.exists("loop")) {
              innerText += capture;
              break;
            }

            if (args.length < 1) {
              throw new Error("No arguments supplied for if statement");
            }
            else if (args.length === 1) {
              if (typeof context[args[1]] !== "number" && (context[args[0]] == 0 || context[args[0]] == null)) {
                mode.set("else");
                mode.set("skip");
                break;
              }
            }
            else {
              if(!runConditional(self, args, context)) {
                mode.set("skip");
              }
            }
          }
          else if (method === "else") {
            if (mode.exists("else")) {
              mode.unset("skip");
            }
            else {
              if(!mode.exists("skip")) {
                mode.set("skip");
              }
            }
          }
          else if (method === "endif") {
            if (mode.exists("loop")) {
              innerText += capture;
              break;
            }

            mode.unset("skip"); 
          }
          else if (method === "each") {
            if (mode.exists("skip")) {
              break;
            }

            if (mode.exists("loop")) {
              mode.set("loop");
              innerText += delimiters.START_EXPR + capture;
              break;
            }

            mode.set("loop");

            if (args.length < 1) {
              throw new Error("No arguments supplied for each statement");
            }
            // Iterating an array
            else if (Array.isArray(context[args[0]])) {
              names = {};

              obj = context[args[0]];
              innerText = '';
            }
            // Iterating an object
            else {
              names = {};

              if (context[args[0]]) {
                obj = context[args[0]];
                names['obj'] = args[0];
                names['key'] = args[2];
                names['val'] = args[3];
                innerText = '';
              }
            }
          }
          else if (method === "endeach") {
            mode.unset("loop");
            if (!mode.exists("loop")) {
              if (Array.isArray(obj)) {
                loop.execArray(self, obj, context);
              }
              else {
                loop.execObj(self, obj, names, context);
              }

              loop.clear();
            }
            else {
              if (mode.exists("loop")) {
                innerText += capture + delimiters.END_EXPR;
                break;
              }
            }
          }
          else if (method === "partial") {
            if (mode.exists("skip")) {
              break;
            }

            if (mode.exists("loop")) {
              innerText += capture;
              break;
            }

            tmpVar = capture.split(' ');
            method = tmpVar.shift();
            args = tmpVar;

            obj = runPartial(self, method, args, obj);
            mode.set("partial");
            mode.set("expr");
          }
        }
        else {
          if (mode.exists("skip")) {
            break;
          }

          if (mode.exists("loop")) {
            innerText += capture;
            break;
          }

          output += capture;
        }

        break;
      
      case "WHITESPACE":
        if (mode.exists("skip")) {
          break;
        }

        if (mode.exists("loop")) {
          innerText += capture;
          break;
        }

        output += capture;

        break;

      case "FILTER":
        if (mode.exists("skip")) {
          break;
        }

        if (mode.exists("loop")) {
          innerText += capture;
          break;
        }

        mode.unset("prop");
        mode.set("filter");

        break;

      case "COMMENT":
        if (mode.exists("skip")) {
          obj = "--";
          break;
        }

        if (mode.exists("loop")) {
          innerText += capture;
          break;
        }

        if (mode.exists("expr")) {
          mode.set("skip");
        }
        else {
          output += capture;
        }

        break;
    }

    if (stack.length) {
      return parse(self, context, stack, delimiters);
    }

    // Store output before resetting scope
    retVal = output;

    // Reset outer scope
    mode.clear();
    output = '';
    obj = undefined;

    // Return value
    return retVal;
  };
}();

function main(self, template, context, delimiters) {
  var string, sp, ep, se, ee, co, fi;
  var error = 0;
  var stack = [];
  var grammar = {};

  sp = "\\" + delimiters.START_PROP.split("").join("\\");
  ep = "\\" + delimiters.END_PROP.split("").join("\\");
  se = "\\" + delimiters.START_EXPR.split("").join("\\");
  ee = "\\" + delimiters.END_EXPR.split("").join("\\");
  co = "\\" + delimiters.COMMENT.split("").join("\\");
  fi = "\\" + delimiters.FILTER.split("").join("\\");

  grammar.START_PROP = RegExp("^" + sp);
  grammar.END_PROP = RegExp("^" + ep);
  grammar.START_EXPR = RegExp("^" + se);
  grammar.END_EXPR = RegExp("^" + ee);
  grammar.COMMENT = RegExp("^" + co);
  grammar.FILTER = RegExp("^" + fi);

  string = [ sp, ep, se, ee, co, fi ].join("|");

  grammar.WHITESPACE = /^[\ |\t|\r|\n]+/,
  grammar.STRING = RegExp("^((?!" + string + ").)*");

  if (error = tokenizer(template, stack, grammar)) {
    throw new Error(error);
  }

  stack.reverse();

  return parse(self, context, stack, delimiters);
}

module.exports = main;
